#!/usr/bin/env python
"""Create, delete, or change a group of Google Compute Engine VMs with names
like "{prefix}-0", "{prefix}-1", "{prefix}-2", ...

# Example: Create worker VMs grace-wcm-0, grace-wcm-1, grace-wcm-2 with metadata
# db=analysis so those workers will use the named database.
    gce grace-wcm -c3 -m db=analysis

# Example: Delete those 3 worker VMs.
    gce --delete grace-wcm -c3 -d

# Example: Set their metadata field `quit` to `when-idle`, asking Fireworkers to
# shut down when idle.
    gce grace-wcm -c3 --set -m quit=when-idle
"""

from __future__ import absolute_import, division, print_function

import argparse
import os
from pprint import pprint
import re
import sys

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

from typing import Any, Dict, List, Optional
from borealis.util import gcp

#: Access Scopes for the created GCE VMs.
SCOPES = ','.join([
    'compute-rw',           # for the instance to delete itself when done
    'logging-write',        # to write StackDriver logs
    'storage-rw',           # to read/write GCS files
    'monitoring-write',     # to emit data for load monitoring
    'service-control',      # ?
    'service-management',   # ?
    'trace',                # for debugging
])


def _clean(token):
    # type: (Any) -> str
    """Clean the token of "=" and "," chars so it won't mess up a
    "key=val,key=val" metadata string.
    """
    return re.sub(r'[=,]+', '', str(token))


def _join_metadata(metadata):
    # type: (Dict[str, Any]) -> str
    """Join the metadata dictionary into a suitable shell token string."""
    return ','.join(
        '{}={}'.format(_clean(k), _clean(v)) for k, v in metadata.items())


def _options_list(options):
    # type: (dict) -> List[str]
    """Translate a dict into a list of CLI "--option=value" tokens."""
    # E.g. ["--description=fire worker", "--metadata=k1=v1,k2=v2", "--quiet"]
    return ['--{}{}'.format(_clean(k), '' if v is None else '={}'.format(v))
            for k, v in options.items()]


class ComputeEngine(object):
    """Runs `gcloud compute` to create, delete, or change a group of GCE VM
    instances named "{prefix}-{index}".
    """

    MAX_VMS = 100  # don't create more than this many GCE VMs at a time

    def __init__(self, name_prefix, dry_run=False, verbose=False):
        # type: (str, bool, bool) -> None
        assert name_prefix, 'the name_prefix must not be empty'

        self.name_prefix = name_prefix
        self.dry_run = dry_run
        self.verbose = verbose

    def make_names(self, base=0, count=1):
        # type: (int, int) -> List[str]
        """Return a list of GCE VM names in the form "{prefix}-{index}" over the
        range [base .. count], sanitizing the name prefix to a legal VM name
        string and eliding "workflow" for brevity.
        """
        sanitized = re.sub(
            r'[^-a-z0-9]+', '-',
            self.name_prefix.lower().replace('workflow', ''))
        names = ['{}-{}'.format(sanitized, i) for i in range(base, base + count)]
        return names

    def _log_header(self, action, instance_names):
        # type: (str, List[str]) -> None
        dry = 'Dry run for: ' if self.dry_run else ''
        count = len(instance_names)
        vms = ('VMs' if count == 0
               else 'VM: {}'.format(instance_names[0]) if count == 1
               else 'VMs: {} .. {}'.format(instance_names[0], instance_names[-1]))
        print('{}{} {} Google Compute Engine {}'.format(dry, action, count, vms))

    def create(self, base=0, count=1, command_options=None, **metadata):
        # type: (int, int, Optional[Dict[str, Any]], **Any) -> None
        """In parallel, create a group of GCE VM instances.

        This provides default command options to `gcloud compute instances create`.
        The caller should at least set the `image-family` option and override
        default options as needed.

        This converts `command_options` and `metadata` to strings, stripping any
        ',' and '=' characters from the metadata fields, then passes tokens to
        `gcloud` without shell quoting risks.

        If `dry_run`, this logs the constructed `gcloud` command instead of
        running it, or if `verbose`, this logs the `gcloud` command before
        running it.
        """
        assert 0 <= count <= self.MAX_VMS, 'create-instance count ({}) must be in the range [0 .. {}]'.format(
            count, self.MAX_VMS)
        instance_names = self.make_names(base, count)

        self._log_header('Creating', instance_names)
        if count <= 0:
            return

        project = gcp.project()
        options = {
            'project': project,
            'image-project': project,
            'zone': gcp.zone(),
            'machine-type': 'n1-standard-1',
            'subnet': 'default',
            'network-tier': 'PREMIUM',
            'maintenance-policy': 'MIGRATE',
            'boot-disk-type': 'pd-standard',
            'scopes': SCOPES,
        }
        options.update(command_options or {})

        metadata_string = _join_metadata(metadata)
        if metadata_string:
            options['metadata'] = metadata_string

        options_list = _options_list(options)
        cmd_tokens = ['gcloud', 'compute', 'instances', 'create'
            ] + instance_names + options_list

        if self.dry_run or self.verbose:
            pprint(cmd_tokens)

        if not self.dry_run:
            subprocess.call(cmd_tokens)

    def delete(self, base=0, count=1, command_options=None):
        # type: (int, int, Optional[Dict[str, Any]]) -> None
        """In parallel, delete a group of GCE VM instances.

        If `dry_run`, this logs the constructed `gcloud` command instead of
        running it, or if `verbose`, this logs the `gcloud` command before
        running it.

        If command_options includes {'quiet': None}, gcloud won't ask for
        confirmation to irreversibly auto-delete disks.
        """
        instance_names = self.make_names(base, count)

        self._log_header('Deleting', instance_names)
        if count <= 0:
            return

        project = gcp.project()
        options = {
            'project': project,
            'zone': gcp.zone()}
        options.update(command_options or {})

        options_list = _options_list(options)
        cmd_tokens = ['gcloud', 'compute', 'instances', 'delete'
            ] + instance_names + options_list

        if self.dry_run or self.verbose:
            pprint(cmd_tokens)

        if not self.dry_run:
            subprocess.call(cmd_tokens)

    def set_metadata(self, base=0, count=1, command_options=None, **metadata):
        # type: (int, int, Optional[Dict[str, Any]], **Any) -> None
        """Set metadata fields on a group of GCE VM instances.

        If `dry_run`, this logs the constructed `gcloud` command instead of
        running it, or if `verbose`, this logs the `gcloud` command before
        running it.

        NOTE: This supports 'key=value' and 'key=' but doesn't yet support a
        plain 'key' which is the way to remove a key/value field.
        """
        instance_names = self.make_names(base, count)

        self._log_header('Setting metadata on', instance_names)
        if count <= 0:
            return

        project = gcp.project()
        metadata_string = _join_metadata(metadata)
        options = {
            'project': project,
            'zone': gcp.zone(),
            'metadata': metadata_string}
        options.update(command_options or {})

        options_list = _options_list(options)
        for name in instance_names:
            cmd_tokens = ['gcloud', 'compute', 'instances', 'add-metadata', name
                ] + options_list

            if self.dry_run or self.verbose:
                pprint(cmd_tokens)

            if not self.dry_run:
                subprocess.call(cmd_tokens)


def cli():
    parser = argparse.ArgumentParser(
        description='''Create, delete, or set metadata on a group of Google
            Compute Engine VMs, e.g. workflow workers that start up from a disk
            image. (This code also has an API for direct use.)''')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--delete', action='store_const', dest='action',
        const='delete', default='create',
        help='Delete VMs instead of creating VMs.')
    group.add_argument('--set-metadata', action='store_const', dest='action',
        const='metadata',
        help='Set metadata on VMs instead of creating VMs. Set `-m quit=when-idle`'
             ' to ask the specified Fireworkers to shut down gracefully.')

    parser.add_argument('name_prefix', metavar='NAME-PREFIX',
        help='The GCE VM name prefix for constructing a batch of VM names of the'
             ' form {PREFIX}-{NUMBER}.')
    parser.add_argument('-d', '--dry-run', action='store_true', dest='dry_run',
        help='Dry run: Print the constructed gcloud commands and exit.')
    parser.add_argument('-b', '--base', type=int, default=0,
        help='The base number for the VM name suffixes (default=0). This is'
             ' mainly for creating additional VMs since they need unique names.')
    parser.add_argument('-c', '--count', type=int, default=1,
        help='The number of VMs to create/delete/set (default=1).')
    parser.add_argument('-f', '--family', default='fireworker',
        help='The GCE Disk Image Family to create from (default="fireworker").'
             ' ("sisyphus-worker" is also interesting.) (With "fireworker", be'
             ' sure to set `-m db=MY_DATABASE_NAME`. With "sisyphus-worker", be'
             ' sure to set `-m workflow=MY_WORKFLOW_NAME`.)')
    parser.add_argument('-s', '--service-account', dest='service_account',
        help='The service account identity to attach when creating VMs, e.g.'
             ' "999999999999-compute@developer.gserviceaccount.com".')
    parser.add_argument('-m', '--metadata', metavar='KEY=VALUE',
        action='append', default=[],
        help='A custom GCE metadata "key=value" setting, e.g. "db=analyze" to'
             ' point FireWorks workers to a LaunchPad database. (You can use'
             ' this option zero or more times.)')

    args = parser.parse_args()
    unpacked = [e.split('=', 2) + [''] for e in args.metadata]
    metadata = {e[0]: e[1] for e in unpacked}

    # Cross-check the args.
    if args.action == 'create':
        if args.family == 'sisyphus-worker':
            assert metadata.get('workflow'), (
                'need `-m workflow=MY_WORKFLOW_NAME` to create sisyphus workers')
        if args.family == 'fireworker':
            assert metadata.get('db'), (
                'need `-m db=MY_DATABASE_NAME` to create Fireworkers')
    elif args.action == 'metadata':
        assert metadata, 'need some metadata to set'

    compute_engine = ComputeEngine(args.name_prefix, dry_run=args.dry_run)

    if args.action == 'create':
        options = {
            'image-family': args.family,
            'description': args.family + ' worker'}
        if args.service_account:
            options['service-account'] = args.service_account
        compute_engine.create(args.base, args.count, command_options=options,
                              **metadata)
    elif args.action == 'delete':
        compute_engine.delete(args.base, args.count)
    elif args.action == 'metadata':
        compute_engine.set_metadata(args.base, args.count, **metadata)


if __name__ == '__main__':
    cli()



"""A Firetask that runs a shell command in a Docker container, pulling input
files from Google Cloud Storage (GCS) and pushing output files to GCS.
"""

from __future__ import absolute_import, division, print_function

from collections import namedtuple
import logging
import os
from pprint import pprint
import shutil
from typing import Any, List, Optional

import docker
from docker import errors as docker_errors
from docker.types import Mount
from docker.utils import parse_repository_tag
from fireworks import explicit_serialize, FiretaskBase, FWAction

import borealis.util.filepath as fp
import borealis.util.storage as st


class DockerTaskError(Exception):
    pass


PathMapping = namedtuple('PathMapping', 'captures local_prefix local sub_path mount')


def uid_gid():
    """Return the Unix uid:gid (user ID, group ID) pair."""
    return '{}:{}'.format(fp.run_cmdline('id -u'), fp.run_cmdline('id -g'))


def captures(path):
    # type: (str) -> Optional[str]
    """Categorize the given output path as capturing a log (>>), capturing
    just stdout + stderr (>), or neither (a regular file or directory).
    """
    return ('>>' if path.startswith('>>')
            else '>' if path.startswith('>')
            else None)


@explicit_serialize
class DockerTask(FiretaskBase):
    _fw_name = 'DockerTask'

    # Parms
    # -----
    # name: the payload task name, for logging.
    #
    # image: the Docker Image to pull.
    #
    # command: the shell command tokens to run in the Docker Container.
    #
    # internal_prefix: the base pathname inside the Docker Container for inputs
    #   and outputs (files and directory trees).
    #
    # storage_prefix: the GCS base pathname for inputs and outputs.
    #
    # inputs, outputs: absolute pathnames internal to the Docker container of
    #   input/output files and directories to pull/push to GCS. DockerTask will
    #   construct the corresponding GCS storage paths by rebasing each path from
    #   `internal_prefix` to `storage_prefix`, and the corresponding local paths
    #   outside the container by rebasing each path to `local_prefix`.
    #
    #   NOTE: Each path in `inputs` and `outputs` names a directory tree of
    #   files if it ends with '/', otherwise a file. This is needed because the
    #   GCS is a flat object store without directories and DockerTask needs to
    #   know whether to create files or directories.
    #
    #   If the task completes normally, DockerTask will all its outputs to GCS.
    #   Otherwise, it only writes '>>' log files.
    #
    #   An output path that starts with '>>' will capture a log of stdout +
    #   stderr + other log messages like elapsed time and task exit code.
    #   DockerTask will write it even if the task fails (to aid debugging),
    #   unlike all the other outputs.
    #
    #   An output path that starts with '>' captures stdout + stderr. The rest
    #   of the path is as if internal to the container, and will get rebased to
    #   to compute its storage path.
    #
    # timeout: in seconds, indicates how long to let the task run.
    required_params = [
        'name',
        'image',
        'command',
        'internal_prefix',
        'storage_prefix']
    optional_params = [
        'inputs',
        'outputs',
        'timeout']

    LOCAL_BASEDIR = os.path.join(os.sep, 'tmp', 'fireworker')

    def _log(self):
        # type: () -> logging.Logger
        """Return a Logger for this task."""
        parent = logging.getLogger('dockerfiretask')
        parent.setLevel(logging.DEBUG)

        name = 'dockerfiretask.{}'.format(self['name'])
        return logging.getLogger(name)

    def pull_docker_image(self, docker_client):
        # type: (docker.DockerClient) -> Any  # a Docker Image
        """Pull the requested Docker Image. Ensure there's a tag so pull() will
        get one Image rather than all tags in a repository.
        """
        repository, tag = parse_repository_tag(self['image'])
        if not tag:
            tag = 'latest'  # 'latest' is the default tag; it doesn't mean squat
        self._log().info('Pulling Docker image %s:%s', repository, tag)
        return docker_client.images.pull(repository, tag)

    def rebase(self, internal_path, new_prefix):
        # type: (str, str) -> str
        """Strip off any '>' or '>>' prefix then rebase the internal-to-container
        path from the internal_prefix to new_prefix.

        '>' or '>>' means capture stdout + stderr rather than fetching a
        container file; the rest of internal_path is the as-if-internal path.
        '>>' creates a log of stdout + stderr + additional information and
        writes it even if the Docker command fails.

        A path ending with '/' indicates a directory.
        """
        core_path = internal_path.lstrip('>')
        internal_prefix = self['internal_prefix']
        rel_path = st.relpath(core_path, internal_prefix)
        new_path = os.path.join(new_prefix, rel_path)

        if '..' in new_path:
            # This could happen if `internal_path` doesn't start with
            # `internal_prefix`.
            raise ValueError('Rebased path "{}" contains ".."'.format(new_path))
        return new_path

    def setup_mount(self, internal_path, local_prefix):
        # type: (str, str) -> PathMapping
        """Create a PathMapping between a path internal to the Docker container
        and a sub_path relative the storage_prefix (GCS) and the local_prefix
        (local file system), make the Docker local:internal Mount object, and
        create the local file or directory so Docker can detect whether to mount
        a file or directory.
        """
        local_path = self.rebase(internal_path, local_prefix)
        sub_path = self.rebase(internal_path, '')

        fp.makedirs(os.path.dirname(local_path))
        if not st.names_a_directory(local_path):
            open(local_path, 'a').close()

        # Create the Docker Mount unless this mapping will capture stdout & stderr.
        caps = captures(internal_path)
        mount = (None if caps
                 else Mount(target=internal_path, source=local_path, type='bind'))

        return PathMapping(caps, local_prefix, local_path, sub_path, mount)

    def setup_mounts(self, group):
        # type: (str) -> List[PathMapping]
        """Set up all the mounts for the 'inputs' or 'outputs' group."""
        return [self.setup_mount(path, os.path.join(self.LOCAL_BASEDIR, group))
                for path in self.get(group, [])]

    def _outputs_to_push(self, lines, success, outs, epilogue):
        # type: (List[str], bool, List[PathMapping], str) -> List[PathMapping]
        """Write requested stdout+stderr and log output files, then return a
        list of all output PathMappings to push to GCS. That's all of them if
        the Task succeeded; only the '>>' logs if it failed.
        """
        to_push = []

        for out in outs:
            if out.captures:
                try:
                    with open(out.local, 'w') as f:
                        hr = ''
                        if out.captures == '>>':
                            hr = '-' * 80
                            pprint(self.to_dict(), f)  # prologue
                            f.write('\n{}\n'.format(hr))

                        f.writelines(lines)

                        if hr:
                            f.write('{}\n\n{}\n'.format(hr, epilogue))
                except IOError as e:
                    self._log().exception('Error capturing to %s', out.local)

            if success or out.captures == '>>':
                to_push.append(out)

        return to_push

    def push_to_gcs(self, to_push):
        # type: (List[PathMapping]) -> bool
        """Push outputs to GCS. Return True if successful."""
        ok = True
        prefix = self['storage_prefix']

        self._log().info('Pushing %s outputs to GCS %s: %s',
            len(to_push), prefix, [mapping.sub_path for mapping in to_push])
        gcs = st.CloudStorage(prefix)

        for mapping in to_push:
           ok = gcs.upload_tree(mapping.local, mapping.sub_path) and ok

        return ok

    def pull_from_gcs(self, to_pull):
        # type: (List[PathMapping]) -> bool
        """Pull inputs from GCS. Return True if successful."""
        ok = True
        prefix = self['storage_prefix']

        self._log().info('Pulling %s inputs from GCS %s: %s',
            len(to_pull), prefix, [mapping.sub_path for mapping in to_pull])
        gcs = st.CloudStorage(prefix)

        for mapping in to_pull:
            ok = gcs.download_tree(mapping.sub_path, mapping.local_prefix) and ok

        return ok


    # ODDITIES ABOUT THE PYTHON DOCKER PACKAGE
    #
    # images.pull() will pull a list of images if neither arg gives a tag. We
    # don't want that, so provide a default tag if the image name doesn't
    # include one.
    #
    # Some variations on the call sequence bury the exit code and stdout, so
    # test carefully when changing it and remove the container manually.
    #
    # Mounts need to be set up just right. Use type='bind' and pre-create the
    # local file or directory.
    #
    # Example container.attrs['State'] (if it is a dict):
    #   {'Dead': False,
    #    'Error': '',
    #    'ExitCode': 0,        # always seems to be 0!
    #    'FinishedAt': '0001-01-01T00:00:00Z',
    #    'OOMKilled': False,
    #    'Paused': False,
    #    'Pid': 0,
    #    'Restarting': False,
    #    'Running': False,
    #    'StartedAt': '0001-01-01T00:00:00Z',
    #    'Status': 'created'}  # always seems to be 'created'!
    #
    # Examples of container.wait() results:
    #   {'Error': None, 'StatusCode': 0}
    #   {'Error': None, 'StatusCode': 1}

    def run_task(self, fw_spec):
        # type: (dict) -> Optional[FWAction]
        """Run a task as a shell command in a Docker container."""
        name = self['name']
        errors = []  # type: List[str]
        lines = []  # type: List[str]
        logger = self._log()

        def check(success, or_error):
            if not success:
                errors.append(or_error)

        def epilog():
            # TODO(jerry): Include the elapsed time and the timeout parameter.
            return '{} task: {} {}'.format(
                'FAILED' if errors else 'SUCCESSFUL', name, errors if errors else '')

        logger.warning('STARTING TASK: %s', name)

        try:
            docker_client = docker.from_env()
            image = self.pull_docker_image(docker_client)

            ins = self.setup_mounts('inputs')
            outs = self.setup_mounts('outputs')

            check(self.pull_from_gcs(ins), 'Failed to pull inputs')

            logger.info('Running: %s', self['command'])
            container = docker_client.containers.run(
                image,
                command=self['command'],
                user=uid_gid(),
                mounts=[mapping.mount for mapping in ins + outs if mapping.mount],
                detach=True)

            try:
                for line in container.logs(stream=True):  # TODO: Set follow=True?
                    # NOTE: Can call stream.close() to cancel from another thread.
                    line = line.decode()
                    lines.append(line)
                    logger.info('%s', line.rstrip())

                exit_code = container.wait()['StatusCode']
                check(exit_code == 0, 'Exit code {}{}'.format(
                    exit_code, ' (SIGKILL)' if exit_code == 137 else ''))
                state = container.attrs.get('State')
                if isinstance(state, dict):
                    check(not state.get('OOMKilled'), 'OOM-Killed')
            finally:
                try:
                    container.remove(force=True)
                except docker_errors.APIError as e:  # troubling but not a task error
                    logger.exception('Docker error removing a container')

            to_push = self._outputs_to_push(lines, not errors, outs, epilog())

            # NOTE: The >>task.log file won't report push failures since it's
            # written before pushing and might itself fail to push. But the
            # StackDriver log will get it, and the Fireworks stored_data will
            # get it if this returns FWAction rather than raise an exception.
            check(self.push_to_gcs(to_push), 'Failed to push outputs')

        except (Exception, KeyboardInterrupt) as e:
            # Log it, clean up, and re-raise it. That'll FIZZLE the Firework.
            check(False, repr(e))
            raise
        finally:
            logger.warning('%s', epilog())

            # [Could wipe just os.path.join(self.LOCAL_BASEDIR, 'inputs') to
            # keep the outputs for local scrutiny.]
            wipe_out = self.LOCAL_BASEDIR
            shutil.rmtree(wipe_out, ignore_errors=True)

        if errors:
            raise DockerTaskError(repr(errors))  # FIZZLE this Firework

        return None



#!/usr/bin/env python
"""A Fireworks worker on Google Compute Engine to "rapidfire" launch rockets.

    python -m borealis.fireworker

The borealis-fireworker installs a console_scripts for fireworker and gce, so
you can simply run

    fireworker

NOTE: When running as a systemd service or otherwise outside an interactive
console, set the `PYTHONUNBUFFERED=1` environment variable or run with
`python -u fireworker.py` so the logging output comes out in real time rather
than buffering up into long delayed chunks.
"""

from __future__ import absolute_import, division, print_function

import argparse
import logging
import socket
import sys
import time
from typing import Any, Dict

from fireworks import LaunchPad, FWorker, fw_config
from fireworks.core import rocket_launcher
import google.cloud.logging as gcl
from google.cloud.logging.resource import Resource
import ruamel.yaml as yaml

# import logging_tree

from borealis.util import gcp
from borealis.util.log_filter import LogPrefixFilter

#: The standard launchpad config filename (in CWD) to read.
#: GCE instance metadata will override some field values.
LAUNCHPAD_YAML = 'gce_my_launchpad.yaml'
LAUNCHPAD_YAML2 = 'my_launchpad.yaml'
DEFAULT_FIREWORKS_DATABASE = 'default_fireworks_database'

ERROR_EXIT_CODE = 1
KEYBOARD_INTERRUPT_EXIT_CODE = 2

#: Fireworker logger.
FW_LOGGER = logging.getLogger('fireworker')
FW_LOGGER.setLevel(logging.DEBUG)

#: Fireworker console-only logger.
FW_CONSOLE_LOGGER = logging.getLogger('fireworker.console')
FW_CONSOLE_LOGGER.setLevel(logging.DEBUG)
FW_CONSOLE_LOGGER.propagate = False


def _setup_logging(gce_instance_name, host_name):
    # type: (str, str) -> None
    """Set up GCP StackDriver cloud logging on Python's root logger for the GCE
    instance name or any host name. Set a narrow logging filter if running off
    GCE (instance_name is empty).
    """
    exclude = (FW_CONSOLE_LOGGER.name, 'urllib3')

    monitored_resource = Resource(
        type='gce_instance',
        labels={  # Add a 'tag' label? It gets 'project_id' automatically.
            'instance_id': host_name,
            'zone': gcp.zone()})
    client = gcl.Client()

    # noinspection PyTypeChecker
    client.setup_logging(
        log_level=logging.WARNING,
        excluded_loggers=exclude,
        name=FW_LOGGER.name,
        resource=monitored_resource)

    # For aggregate cloud logs: Filter out debug details and when running off
    # GCE also filter out the dockerfiretask payload stdout lines, keeping just
    # the WARNINGs and start/end messages. (Use WARNING rather than NOTICE level
    # for start/end because Logs Viewer doesn't support NOTICE very well.) We
    # can't `exclude` those loggers since that'd block their WARNINGs.
    #
    # Console logs: Filter out messages already printed by handlers of nested
    # loggers "launchpad" and "rocket.launcher", allowing WARNINGs just in case.
    root = logging.getLogger()
    fworker_level = logging.DEBUG if gce_instance_name else logging.WARNING
    cloud_filter = LogPrefixFilter(
        {'fireworker': fworker_level, 'dockerfiretask': fworker_level},
        logging.WARNING)
    console_filter = LogPrefixFilter(
        {'fireworker': logging.INFO, 'dockerfiretask': logging.DEBUG},
        logging.WARNING)
    for handler in root.handlers:
        # This `is_cloud` test is a bit fragile.
        is_cloud = hasattr(handler, 'transport') or hasattr(handler, 'resource')
        handler.addFilter(cloud_filter if is_cloud else console_filter)


def _cleanup_logging():
    # type: () -> None
    """Clean up StackDriver cloud logging: Flush and remove root logger's
    background-transport handlers so the last messages get to the server and
    won't raise RuntimeError('cannot schedule new futures after shutdown').

    StackDriver should be out of the loop after this but there's no documented
    API for this so hopefully it's right, idempotent, and safe if StackDriver
    logging was not set up.
    """
    root = logging.getLogger()

    for handler in root.handlers.copy():
        if hasattr(handler, 'transport'):
            transport = handler.transport
            if hasattr(transport, 'flush'):
                transport.flush()
                root.removeHandler(handler)


class Fireworker(object):
    """A Fireworks worker on Google Compute Engine to "rapidfire" launch rockets.

    NOTE: When running as a systemd service or otherwise outside an interactive
    console, set the `PYTHONUNBUFFERED=1` environment variable or run with
    `python -u fireworker.py` so the logging output comes out in real time rather
    than buffering up into long delayed chunks.
    """

    def __init__(self, lpad_config, host_name):
        # type: (Dict[str, Any], str) -> None
        self.lpad_config = lpad_config
        self.host_name = host_name

        # NOTE: FireWorks creates loggers with stdout stream handlers for each
        # (name, level) pair. So setting strm_lvl='WARNING' gets both INFO and
        # WARNING handlers which might print duplicate lines. Try to tame it.
        self.strm_lvl = lpad_config.get('strm_lvl') or 'INFO'
        fw_config.ROCKET_STREAM_LOGLEVEL = self.strm_lvl

        self.sleep_secs = 10
        self.idle_for_waiters = 60 * 60
        self.idle_for_queued = 15 * 60  # TODO(jerry): Rename this

        self.launchpad = LaunchPad(**lpad_config)
        self.launchpad.m_logger.setLevel(self.strm_lvl)  # set non-stream level

        # Can optionally set a specific `category` of jobs to pull, a `query`
        # to restrict the type of Fireworks to run, and an `env` to pass
        # worker-specific into to the Firetasks.
        self.fireworker = FWorker(host_name)

    def launch_rockets(self):
        # type: () -> None
        """Keep launching rockets that are ready to go. Stop after:
          * idling idle_for_waiters secs for WAITING rockets to become ready,
          * idling idle_for_queued secs if no rockets are even waiting,
          * the custom metadata field `attributes/quit` becomes 'when-idle'.

        The first timeout should be long enough to wait around to run queued
        rockets after running rockets finish prerequisite work. The second
        timeout should be long enough to let new work get queued.
        """

        # rapidfire() launches READY rockets until: `max_loops` batches of READY
        # rockets OR `timeout` total elapsed seconds OR `nlaunches` rockets launched
        # OR `nlaunches` == 0 ("until completion", the default) AND no rockets are
        # even waiting.
        #
        # Set max_loops so it won't loop forever and we can track idle time.
        #
        # TODO(jerry): Set m_dir? local_redirect?
        while True:
            rocket_launcher.rapidfire(
                self.launchpad, self.fireworker, strm_lvl=self.strm_lvl,
                max_loops=1, sleep_time=self.sleep_secs)

            # logging_tree.printout()  # *** DEBUG ***

            # Idle to the max.
            idled = self.sleep_secs  # rapidfire() just slept once
            while not self.launchpad.run_exists(self.fireworker):  # none ready to run
                future_work = self.launchpad.future_run_exists(self.fireworker)  # any waiting?
                if idled >= (self.idle_for_waiters if future_work else self.idle_for_queued):
                    return

                if gcp.instance_metadata('attributes/quit') == 'when-idle':
                    FW_LOGGER.info('Quitting by "when-idle" request')
                    return

                FW_CONSOLE_LOGGER.info(
                    'Sleeping for %s secs waiting for launchable rockets',
                    self.sleep_secs)
                time.sleep(self.sleep_secs)
                idled += self.sleep_secs


class Redacted(object):
    def __repr__(self):
        """Print without quotes to look like a mask not a lame password."""
        return '*****'


def _read_launchpad_config():
    # type: () -> dict
    """Read the primary LAUNCHPAD_YAML or else the fallback LAUNCHPAD_YAML2.
    Raises FileNotFoundError if neither file was found.
    """
    try:
        with open(LAUNCHPAD_YAML) as f1:
            return yaml.safe_load(f1)  # type: dict
    except FileNotFoundError as e1:

        try:
            with open(LAUNCHPAD_YAML2) as f2:
                return yaml.safe_load(f2)  # type: dict
        except FileNotFoundError as _:
            pass

        raise  # continue propagating e1 (naming the primary file)


def main(development=False):
    # type: (bool) -> None
    """Run as a FireWorks worker node on Google Compute Engine (GCE), launching
    Fireworks rockets in rapidfire mode then deleting this GCE VM instance.

    Get configuration settings from GCE VM metadata fields:
        name - the Fireworker name [required]
        attributes/db - DB name (user-specific or workflow-specific) [required]
        attributes/username - DB username [optional]
        attributes/password - DB password [optional]
    secondarily from a launchpad yaml file:
        host, port - for the DB connection [required]
        logdir, strm_lvl, ... [optional, for "launchpad" & "rocket" logging]
        DB name, DB username, and DB password [fallback]
    with fallbacks:
        name - 'fireworker'
        DB name - DEFAULT_FIREWORKS_DATABASE
        DB username, DB password - null
        logdir, strm_lvl - FireWorks defaults

    The DB username and password are needed if MongoDB is set up to require
    authentication, and it could use shared or user-specific accounts.

    TODO: Add configuration settings for idle_for_waiters and idle_for_queued.

    You can set a custom metadata field to make this worker stop idling:
        gcloud compute instances add-metadata INSTANCE-NAME --metadata quit=when-idle
    """
    exit_code = ERROR_EXIT_CODE

    try:
        instance_name = gcp.gce_instance_name()
        host_name = instance_name or socket.gethostname()
        _setup_logging(instance_name, host_name)

        lpad_config = _read_launchpad_config()

        db_name = (gcp.instance_metadata('attributes/db')
                   or lpad_config.get('name', DEFAULT_FIREWORKS_DATABASE))
        lpad_config['name'] = db_name

        username = (gcp.instance_metadata('attributes/username')
                    or lpad_config.get('username'))
        password = (gcp.instance_metadata('attributes/password')
                    or lpad_config.get('password'))
        lpad_config['username'] = username
        lpad_config['password'] = password

        redacted_config = dict(lpad_config, password=Redacted())
        FW_LOGGER.warning(
            '\nStarting Fireworker on %s with LaunchPad config: %s\n',
            host_name, redacted_config)

        fireworker = Fireworker(lpad_config, host_name)
        fireworker.launch_rockets()

        FW_LOGGER.warning("Fireworker -- normal exit")
        exit_code = 0
    except KeyboardInterrupt:
        FW_LOGGER.warning('Fireworker -- KeyboardInterrupt exit')
        exit_code = KEYBOARD_INTERRUPT_EXIT_CODE
    except Exception:
        FW_LOGGER.exception('Fireworker -- error exit')

    _cleanup_logging()
    _shut_down(development, exit_code)


def _shut_down(development, exit_code):
    # type: (bool, int) -> None
    """Shut down this program or this entire GCE VM (if running on GCE and not
    `development` and `exit_code` isn't KEYBOARD_INTERRUPT_EXIT_CODE).
    """
    if development or exit_code == KEYBOARD_INTERRUPT_EXIT_CODE:
        sys.exit(exit_code)
    else:
        if exit_code:  # an unexpected failure, e.g. missing a needed pip
            FW_CONSOLE_LOGGER.warning(
                'Delaying before deleting this GCE VM to allow some time to'
                ' connect to it and stop this service so you can fix the problem'
                ' and make a new Disk Image.')
            time.sleep(15 * 60)

        gcp.delete_this_vm(exit_code)


def cli():
    """Command Line Interpreter to run a Fireworker."""
    parser = argparse.ArgumentParser(
        description=
            'Run as a FireWorks worker node, launching rockets rapidfire.'
            ' Designed for Google Compute Engine (GCE) and Google Cloud Storage'
            ' (GCS).'
            ' Gets configuration settings from GCE metadata attributes and'
            ' {}, with fallbacks.'
            ' NOTE: This fireworks.py file is at "{}"'.format(
                LAUNCHPAD_YAML, __file__))
    parser.add_argument(
        '--development', action='store_true',
        help="Development mode: When done, just exit Python without deleting"
             " this GCE VM worker instance (if running on GCE).")

    args = parser.parse_args()
    main(development=args.development)


if __name__ == '__main__':
    cli()



"""Interface to Google Cloud Storage (GCS)."""

from __future__ import absolute_import, division, print_function

import itertools
import logging
import os
from typing import Iterator, List, Optional, Union, Set

from google.cloud.storage import Blob, Bucket, Client
from google.cloud.exceptions import GoogleCloudError

import borealis.util.filepath as fp


OCTET_STREAM = 'application/octet-stream'


def bucket_path(pathname):
    # type: (str) -> List[str]
    """Split a GCS pathname like `my_bucket/stuff/file.txt` into bucket and path
    parts `['my_bucket', 'stuff/file.txt']`.
    """
    if pathname.startswith(os.sep):
        pathname = pathname[1:]
    parts = pathname.split(os.sep, 1)
    if len(parts) < 2:
        parts.append('')
    return parts


def names_a_directory(path):
    # type: (str) -> bool
    """Return True if the given path names a directory vs. a file (even on GCS,
    existing or yet to be created) by checking if the path ends with a '/'.
    """
    return path.endswith(os.sep)


def relpath(path, start=None):
    # type: (str, Optional[str]) -> str
    """Return a filepath relative to an optional start directory or the current
    directory -- like os.path.relpath() but keeping path's trailing '/'. This is
    just a path computation. start defaults to os.curdir.
    """
    result = os.path.relpath(path, start)
    if names_a_directory(path):
        result = os.path.join(result, '')
    return result


class CloudStorage(object):
    """A higher level interface to a GCS bucket.

    See https://cloud.google.com/storage/docs/naming about legal bucket and
    blob (aka object) names, and note that bucket names are public and must be
    globally unique.

    To run on Google Compute Engine (GCE), the VM needs access Scopes:
        Storage Read Write (storage-rw)
    and its Service Account needs Permissions:
        Storage Object Admin

    To run off GCE, configure a service account "fireworker" with the above
    permissions, get its private key as a json file:
        PROJECT="$(gcloud config get-value core/project)"
        FIREWORKER_KEY="${HOME}/bin/fireworker.json"
        gcloud iam service-accounts keys create "${FIREWORKER_KEY}" \
            --iam-account "fireworker@${PROJECT}.iam.gserviceaccount.com"
    and append an `export` statement to the shell .profile:
        echo "export GOOGLE_APPLICATION_CREDENTIALS=${FIREWORKER_KEY}" >> ~/.profile
    This will avoid a quota warning and limit.
    """

    #: For efficiency, retrieve just these Blob metadata fields.
    #: https://cloud.google.com/storage/docs/json_api/v1/how-tos/performance
    FIELDS = 'items(bucket,name,id,generation,size),nextPageToken'

    def __init__(self, storage_prefix):
        # type: (Union[Bucket, str]) -> None
        """Construct a GCS accessor with the given storage_prefix, which must
        name a GCS bucket and optionally a base path, e.g.
        'curie-workflows/sim/2020-02-02/'. (It should end with a '/' but will
        work if it doesn't.) All operations are relative to this prefix.

        Raise google.api_core.exceptions.NotFound if the bucket doesn't exist.

        File uploads to GCS will automatically create directory placeholder
        entries, which are empty objects with names ending in '/'. GCS doesn't
        require them but they make gcsfuse-mounted volumes 10x faster (gcsfuse
        without the `--implicit-dirs` option).
        """
        self.bucket_name, self.path_prefix = bucket_path(storage_prefix)
        self.path_prefix = os.path.join(self.path_prefix, '')
        if len(self.bucket_name) < 3:
            # get_bucket() checks that the bucket name is legal and the bucket
            # exists, but it trips over an empty name.
            raise ValueError("Invalid bucket name: '{}'".format(self.bucket_name))

        self.client = Client()
        self.bucket = self.client.get_bucket(self.bucket_name)

        #: A cache of directory placeholders already created or verified.
        self._directory_cache = set()  # type: Set[str]

    def clear_directory_cache(self):
        # type: () -> None
        """Clear the cache of directory placeholder names already created."""
        self._directory_cache = set()

    def list_blobs(self, prefix=''):
        # type: (str) -> Iterator[Blob]
        """List Blobs with the given prefix string (which needn't be a
        "directory" name, and it's relative to the storage_prefix), requesting a
        subset of fields for efficiency. Return a Blob Iterator.
        """
        prefix = os.path.join(self.path_prefix, prefix)
        iterator = self.bucket.list_blobs(prefix=prefix, fields=self.FIELDS)
        return iterator

    def make_dirs(self, sub_path):
        # type: (str) -> None
        """Make sub_path's directory placeholders if they don't exist. E.g. for
        'sim/2020/logs/sim.log', make 'sim/', 'sim/2020/', and 'sim/2020/logs/'.
        See clear_directory_cache().
        """
        parts = os.path.join(self.path_prefix, sub_path).split(os.sep)[:-1]

        for prefix in itertools.accumulate(parts, os.path.join):
            dir_name = os.path.join(prefix, '')

            if dir_name not in self._directory_cache:
                self._directory_cache.add(dir_name)

                blob = self.bucket.blob(dir_name)
                # TODO(jerry): If upload_from_string() took the
                #  `ifGenerationMatch=0` arg that'd create the GCS object if it
                #  doesn't already exist, saving the `exists()` round trip.
                # https://github.com/googleapis/google-cloud-python/issues/10105
                try:
                    if not blob.exists():
                        blob.upload_from_string(b'', content_type=OCTET_STREAM)
                except GoogleCloudError as e:
                    # Failing to create a dir placeholder will affect gcsfuse
                    # mounts but won't break the workflow.
                    logging.exception('Failed to make GCS dir "%s"', dir_name)

    def upload_file(self, local_path, sub_path):
        # type: (str, str) -> bool
        """Upload the file named local_path as (not into) the given GCS sub_path
        (which is relative to the storage_prefix).

        Return True if successful.
        """
        full_path = os.path.join(self.path_prefix, sub_path)

        try:
            self.make_dirs(sub_path)

            blob = self.bucket.blob(full_path)
            blob.upload_from_filename(local_path)  # guesses content_type from the path
        except (GoogleCloudError, FileNotFoundError) as e:
            logging.exception(
                'Failed to upload "%s" as GCS "%s"', local_path, full_path)
            return False
        return True

    def upload_tree(self, local_path, sub_path):
        # type: (str, str) -> bool
        """Upload a file or a directory tree as (not into) the given GCS
        sub_path (which is relative to the storage_prefix).

        Return True if successful.
        """
        if not names_a_directory(sub_path):
            return self.upload_file(local_path, sub_path)

        ok = True
        local_abs = os.path.abspath(local_path)

        for dirpath, dirnames, filenames in os.walk(local_path):
            local_rel_path = os.path.relpath(os.path.abspath(dirpath), local_abs)
            if local_rel_path == '.':
                local_rel_path = ''
            storage_subdir = os.path.join(sub_path, local_rel_path)

            for filename in filenames:
                ok = self.upload_file(
                    os.path.join(dirpath, filename),
                    os.path.join(storage_subdir, filename)) and ok

        return ok

    @classmethod
    def download_blob(cls, blob, local_path):
        # type: (Blob, str) -> bool
        """Download a Blob from GCS as (not into) local_path, making directories
        if needed. `blob` must have its `name` and `bucket` fields set.

        Return True if successful.
        """
        if names_a_directory(local_path):
            fp.makedirs(local_path)
            return True

        fp.makedirs(os.path.dirname(local_path))

        try:
            blob.download_to_filename(local_path)
        except GoogleCloudError as e:
            logging.exception(
                'Failed to download GCS "%s" as "%s"', blob.name, local_path)
            return False

        return True

    def download_file(self, sub_path, local_path):
        # type: (str, str) -> bool
        """Download the GCS file named sub_path (relative to the storage_prefix)
        as (not into) the local_path, making local directories if needed.

        Return True if successful.
        """
        full_path = os.path.join(self.path_prefix, sub_path)
        blob = self.bucket.blob(full_path)
        return self.download_blob(blob, local_path)

    def download_tree(self, sub_path, local_prefix):
        # type: (str, str) -> bool
        """Download all files and directories that begin with the sub_path
        prefix (within the storage_prefix) to their same relative paths in
        local_prefix, making directories if needed.

        Return True if successful.
        """
        if not names_a_directory(sub_path):
            local_path = os.path.join(local_prefix, sub_path)
            return self.download_file(sub_path, local_path)

        ok = True

        for blob in self.list_blobs(sub_path):
            local_rel_path = relpath(blob.name, self.path_prefix)
            path = os.path.join(local_prefix, local_rel_path)
            ok = self.download_blob(blob, path) and ok

        return ok
